@@ -1,6 +1,5 @@
 # TODO: Return unvalidated configuration when there is an exception
 # TODO: Watch multiple folders if not all under parent folder
-# TODO: Kill running containers at exit
 import requests
 import shlex
 import subprocess
@@ -10,7 +9,6 @@ from concurrent.futures import (
     ProcessPoolExecutor, ThreadPoolExecutor, as_completed)
 from datetime import datetime
 from logging import getLogger
-from multiprocessing import Manager, Process, Queue
 from os import environ, getenv, symlink
 from os.path import relpath
 from pathlib import Path
@@ -26,16 +24,17 @@ from ..constants import (
     Error,
     AUTOMATION_PATH,
     AUTOMATION_ROUTE,
+    BATCH_ROUTE,
     DISK_DEBOUNCE_IN_MILLISECONDS,
     DISK_POLL_IN_MILLISECONDS,
     HOST,
     MAXIMUM_PORT,
     MINIMUM_PORT,
-    MODE_CODE_BY_NAME,
-    MODE_NAMES,
     PORT,
+    PORT_ROUTE,
     PROXY_URI,
-    RUN_ROUTE,
+    STEP_CODE_BY_NAME,
+    STEP_NAMES,
     TOKEN_LENGTH)
 from ..exceptions import (
     CrossComputeConfigurationError,
@@ -46,9 +45,8 @@ from ..exceptions import (
     CrossComputeExecutionError)
 from ..macros.iterable import group_by
 from ..macros.security import DictionarySafe
-from .configuration import (
-    get_folder_plus_path,
-    load_configuration)
+from ..settings import multiprocessing_context
+from .configuration import get_folder_plus_path, load_configuration
 from .interface import Automation
 from .server import DiskServer
 from .variable import (
@@ -79,13 +77,13 @@ class DiskAutomation(Automation):
                 future.result()
         return instance
 
-    def run(self, user_environment, with_rebuild=True):
+    def run(self, environment, with_rebuild=True):
         for automation_definition in self.definitions:
             prepare_automation(automation_definition, with_rebuild)
         recurring_definitions = []
         for automation_definition in self.definitions:
             run_automation(
-                automation_definition, user_environment, with_rebuild)
+                automation_definition, environment, with_rebuild)
             if automation_definition.interval_timedelta:
                 recurring_definitions.append(automation_definition)
         if not recurring_definitions:
@@ -96,31 +94,28 @@ class DiskAutomation(Automation):
                 delta = automation_definition.interval_timedelta
                 if datetime.now() > last + delta:
                     run_automation(
-                        automation_definition, user_environment,
-                        with_rebuild=False)
+                        automation_definition, environment, with_rebuild=False)
             sleep(1)
 
     def serve(
-            self, environment, host=HOST, port=PORT, with_refresh=False,
-            with_restart=False, root_uri='', allowed_origins=None,
+            self, environment, host=HOST, port=PORT, with_restart=True,
+            with_prefix=True, with_hidden=True, root_uri='',
+            allowed_origins=None,
             disk_poll_in_milliseconds=DISK_POLL_IN_MILLISECONDS,
             disk_debounce_in_milliseconds=DISK_DEBOUNCE_IN_MILLISECONDS):
-        queue = Queue()
-        with Manager() as manager:
-            infos_by_timestamp = manager.dict()
-            safe = DictionarySafe({}, manager.dict(), TOKEN_LENGTH)
-            server = DiskServer(
-                environment, safe, queue, _work, infos_by_timestamp, host,
-                port, with_refresh, with_restart, root_uri, allowed_origins)
-            configuration = self.configuration
-            if not with_refresh and not with_restart:
-                server.serve(configuration)
-            else:
-                server.watch(
-                    configuration,
-                    disk_poll_in_milliseconds,
-                    disk_debounce_in_milliseconds,
-                    self._reload)
+        queue = multiprocessing_context.Queue()
+        with multiprocessing_context.Manager() as manager:
+            changes = manager.dict()
+            safe = DictionarySafe(manager.dict(), TOKEN_LENGTH)
+            DiskServer(
+                environment, safe, queue, _work, changes, host, port,
+                with_restart, with_prefix, with_hidden, root_uri,
+                allowed_origins,
+            ).watch(
+                self.configuration,
+                disk_poll_in_milliseconds,
+                disk_debounce_in_milliseconds,
+                self._reload)
 
     def _reload(self):
         path = self.path
@@ -187,14 +182,17 @@ class AbstractEngine():
             e.automation_definition = automation_definition
             return_code = e.code
             L.error('%s failed; %s', batch_identifier, e)
+        except KeyboardInterrupt:
+            return_code = Error.COMMAND_INTERRUPTED
+            L.error('%s interrupted')
         else:
             L.info('%s done', batch_identifier)
         return _process_batch(automation_definition, batch_definition, [
             'output', 'log', 'debug',
         ], {'debug': {
+            'source_time': reference_time,
             'execution_time_in_seconds': time() - reference_time,
-            'return_code': return_code,
-        }})
+            'return_code': return_code}})
 
     def prepare(self, automation_definition):
         pass
@@ -203,17 +201,18 @@ class AbstractEngine():
 class UnsafeEngine(AbstractEngine):
 
     def run(self, automation_definition, batch_folder, custom_environment):
-        mode_folder_by_name = _get_mode_folder_by_name(
+        step_folder_by_name = _get_step_folder_by_name(
             automation_definition.folder, batch_folder)
+        script_definitions = automation_definition.script_definitions
         script_environment = _prepare_script_environment(
-            mode_folder_by_name, custom_environment, with_path=True)
-        debug_folder = mode_folder_by_name['debug_folder']
+            step_folder_by_name, custom_environment, with_path=True)
+        debug_folder = step_folder_by_name['debug_folder']
         o_path = debug_folder / 'stdout.txt'
         e_path = debug_folder / 'stderr.txt'
         with open(o_path, 'wt') as o_file, open(e_path, 'w+t') as e_file:
-            for script_definition in automation_definition.script_definitions:
+            for script_definition in script_definitions:
                 return_code = _run_script(
-                    script_definition, mode_folder_by_name,
+                    script_definition, step_folder_by_name,
                     script_environment, o_file, e_file)
         return return_code
 
@@ -232,7 +231,7 @@ class PodmanEngine(AbstractEngine):
         (automation_folder / '.containerignore').write_text(
             CONTAINER_IGNORE_TEXT)
         command_texts = [
-            _.get_command_string().format(**CONTAINER_MODE_FOLDER_BY_NAME)
+            _.get_command_string().format(**CONTAINER_STEP_FOLDER_BY_NAME)
             for _ in automation_definition.script_definitions]
         (automation_folder / CONTAINER_SCRIPT_NAME).write_text(
             '\n'.join([_ + CONTAINER_PIPE_TEXT for _ in command_texts]))
@@ -243,22 +242,32 @@ class PodmanEngine(AbstractEngine):
 
     def run(self, automation_definition, batch_folder, custom_environment):
         automation_folder = automation_definition.folder
+        absolute_batch_folder = automation_folder / batch_folder
         container_id, port_packs = _run_podman_image(
             automation_definition, batch_folder, custom_environment)
-        with _proxy_podman_ports(
-                automation_definition, batch_folder, custom_environment,
-                port_packs):
-            return_code = _run_podman_script(
-                automation_definition, batch_folder, container_id)
+        try:
+            with _proxy_podman_ports(
+                    automation_definition, batch_folder, custom_environment,
+                    port_packs):
+                _copy_datasets_into_podman(container_id, automation_definition)
+                _set_podman_folder_owner(absolute_batch_folder, 1000)
+                return_code = _run_podman_command({'cwd': automation_folder}, [
+                    'exec', '--env-file', CONTAINER_ENV_NAME, container_id,
+                    'bash', CONTAINER_SCRIPT_NAME]).returncode
+        except KeyboardInterrupt:
+            return_code = Error.COMMAND_INTERRUPTED
+        except Exception:
+            raise
+        finally:
+            _set_podman_folder_owner(absolute_batch_folder, 0)
+            _run_podman_command({}, ['kill', container_id])
         if return_code in [126, 127]:
-            error_text = (
-                'permission denied' if return_code == 126 else 'not found')
+            x = 'denied permission' if return_code == 126 else 'not found'
             error = CrossComputeConfigurationError(
-                f'command {error_text} in container; please check script '
-                'definitions')
+                'command %s in container; check script definitions' % x)
             error.code = Error.COMMAND_NOT_RUNNABLE
             raise error
-        elif return_code > 0:
+        elif return_code != 0:
             error_text = (
                 automation_folder / batch_folder / 'debug' / 'stderr.txt'
             ).read_text()
@@ -282,14 +291,14 @@ def get_script_engine(engine_name, with_rebuild=True):
 def update_datasets(automation_definition):
     automation_folder = automation_definition.folder
     for dataset_definition in automation_definition.dataset_definitions:
-        target_path = (automation_folder / dataset_definition.path).resolve()
+        target_path = automation_folder / dataset_definition.path
         target_folder = make_folder(target_path.parent)
         reference_configuration = dataset_definition.reference
         reference_path = get_folder_plus_path(reference_configuration)
         if reference_path:
-            source_path = (automation_folder / reference_path).resolve()
+            source_path = automation_folder / reference_path
             if target_path.is_symlink():
-                if target_path == source_path:
+                if target_path.resolve() == source_path.resolve():
                     continue
                 target_path.unlink()
             elif target_path.exists():
@@ -329,8 +338,7 @@ def run_automation(automation_definition, user_environment, with_rebuild=True):
     return ds
 
 
-def _run_automation_single(
-        automation_definition, run_batch, user_environment):
+def _run_automation_single(automation_definition, run_batch, user_environment):
     ds = []
     for batch_definition in automation_definition.batch_definitions:
         ds.append(run_batch(
@@ -351,18 +359,21 @@ def _run_automation_multiple(
             futures.append(executor.submit(
                 run_batch, automation_definition, batch_definition,
                 user_environment))
-        for future in as_completed(futures):
-            ds.append(future.result())
+        try:
+            for future in as_completed(futures):
+                ds.append(future.result())
+        except KeyboardInterrupt:
+            pass
     return ds
 
 
 def _run_script(
-        script_definition, mode_folder_by_name, script_environment,
+        script_definition, step_folder_by_name, script_environment,
         stdout_file, stderr_file):
     command_string = script_definition.get_command_string()
     if not command_string:
         return
-    command_text = command_string.format(**mode_folder_by_name)
+    command_text = command_string.format(**step_folder_by_name)
     automation_folder = script_definition.automation_folder
     script_folder = script_definition.folder
     command_folder = automation_folder / script_folder
@@ -403,7 +414,7 @@ def _run_podman_image(automation_definition, batch_folder, custom_environment):
     automation_folder = automation_definition.folder
     container_env_path = automation_folder / CONTAINER_ENV_NAME
     script_environment = _prepare_script_environment(
-        CONTAINER_MODE_FOLDER_BY_NAME, custom_environment, with_path=False)
+        CONTAINER_STEP_FOLDER_BY_NAME, custom_environment, with_path=False)
     container_env_path.write_text('\n'.join(
         f'{k}={v}' for k, v in script_environment.items()))
     port_definitions = automation_definition.port_definitions
@@ -417,9 +428,9 @@ def _run_podman_image(automation_definition, batch_folder, custom_environment):
             host_port = find_open_port(
                 minimum_port=MINIMUM_PORT,
                 maximum_port=MAXIMUM_PORT)
-            mode_name = port_definition.mode_name
+            step_name = port_definition.step_name
             container_port = port_definition.number
-            port_packs.append((port_id, host_port, mode_name))
+            port_packs.append((port_id, host_port, step_name))
             command_terms.extend(['-p', f'{host_port}:{container_port}'])
         command_terms.extend([
             '-v', f'{absolute_batch_folder}:/home/user/runs/next:Z'])
@@ -439,53 +450,56 @@ def _proxy_podman_ports(
     origin_uri = custom_environment.get('CROSSCOMPUTE_ORIGIN_URI', '')
     relative_uris = []
     if PROXY_URI:
-        def get_session_uri(host_port, relative_uri):
+        def get_port_uri(target_uri, relative_uri):
             requests.post(PROXY_URI + relative_uri, json={
-                'target': f'http://localhost:{host_port}'})
+                'target': target_uri})
             relative_uris.append(relative_uri)
             return origin_uri + relative_uri
     else:
-        def get_session_uri(host_port, relative_uri):
-            return f'http://localhost:{host_port}'
+        def get_port_uri(target_uri, relative_uri):
+            return target_uri
     automation_uri = AUTOMATION_ROUTE.format(
         automation_slug=automation_definition.slug)
-    run_uri = RUN_ROUTE.format(run_slug=batch_folder.name)
+    batch_uri = BATCH_ROUTE.format(batch_slug=batch_folder.name)
     absolute_batch_folder = automation_definition.folder / batch_folder
-    for port_id, host_port, mode_name in port_packs:
-        port_path = absolute_batch_folder / 'debug' / 'ports.dictionary'
-        mode_code = MODE_CODE_BY_NAME[mode_name]
+    port_uri_by_port_id = {}
+    for port_id, host_port, step_name in port_packs:
+        step_code = STEP_CODE_BY_NAME[step_name]
         variable_id = port_id
-        relative_uri = (
-            f'/sessions{automation_uri}{run_uri}/{mode_code}'
-            f'/{variable_id}')
-        session_uri = get_session_uri(host_port, relative_uri)
-        update_variable_data(port_path, {port_id: session_uri})
-    yield
-    for relative_uri in relative_uris:
-        requests.delete(PROXY_URI + relative_uri)
+        relative_uri = PORT_ROUTE.format(
+            uri=f'{automation_uri}{batch_uri}/{step_code}/{variable_id}')
+        port_uri_by_port_id[port_id] = get_port_uri(
+            f'http://localhost:{host_port}', relative_uri)
+    update_variable_data(
+        absolute_batch_folder / 'debug' / 'ports.dictionary',
+        port_uri_by_port_id)
+    try:
+        yield
+    except KeyboardInterrupt:
+        pass
+    except Exception as e:
+        L.exception(e)
+    finally:
+        for relative_uri in relative_uris:
+            requests.delete(PROXY_URI + relative_uri)
 
 
-def _run_podman_script(automation_definition, batch_folder, container_id):
+def _copy_datasets_into_podman(container_id, automation_definition):
     automation_folder = automation_definition.folder
-    absolute_batch_folder = automation_folder / batch_folder
-
     for dataset_definition in automation_definition.dataset_definitions:
         dataset_path = dataset_definition.path
-        _run_podman_command({}, [
-            'exec', container_id, 'mkdir', dataset_path.parent, '-p'])
+        _make_podman_folder(container_id, dataset_path.parent)
         _run_podman_command({'cwd': automation_folder}, [
             'cp', dataset_path, f'{container_id}:{dataset_path}'])
 
-    _run_podman_command({}, [
-        'unshare', 'chown', '1000:1000', str(absolute_batch_folder), '-R'])
-    process = _run_podman_command({'cwd': automation_folder}, [
-        'exec', '--env-file', CONTAINER_ENV_NAME, container_id,
-        'bash', CONTAINER_SCRIPT_NAME])
-    _run_podman_command({}, [
-        'unshare', 'chown', '0:0', str(absolute_batch_folder), '-R'])
 
-    _run_podman_command({}, ['kill', container_id])
-    return process.returncode
+def _make_podman_folder(container_id, folder):
+    _run_podman_command({}, ['exec', container_id, 'mkdir', folder, '-p'])
+
+
+def _set_podman_folder_owner(folder, user_id):
+    _run_podman_command({}, [
+        'unshare', 'chown', f'{user_id}:{user_id}', str(folder), '-R'])
 
 
 def _run_podman_command(options, terms):
@@ -497,7 +511,7 @@ def _run_podman_command(options, terms):
 def _work(automation_queue):
     try:
         while automation_pack := automation_queue.get():
-            worker_process = Process(
+            worker_process = multiprocessing_context.Process(
                 target=_work_one, args=(automation_pack,), daemon=True)
             worker_process.start()
     except KeyboardInterrupt:
@@ -510,11 +524,10 @@ def _work_one(automation_pack):
         engine = get_script_engine(
             automation_definition.engine_name, with_rebuild=False)
         update_datasets(automation_definition)
-        try:
-            engine.run_batch(*automation_pack)
-        except CrossComputeError as e:
-            e.automation_definition = automation_definition
-            L.error(e)
+        engine.run_batch(*automation_pack)
+    except CrossComputeError as e:
+        e.automation_definition = automation_definition
+        L.error(e)
     except KeyboardInterrupt:
         pass
 
@@ -529,29 +542,29 @@ def _prepare_batch(automation_definition, batch_definition):
         automation_definition,
         variable_definitions_by_path.pop('ENVIRONMENT', []),
         data_by_id)
-    mode_folder_by_name = _make_mode_folder_by_name(
+    step_folder_by_name = _make_step_folder_by_name(
         automation_definition.folder, batch_folder)
     if not data_by_id:
         return batch_folder, batch_environment
-    input_folder = mode_folder_by_name['input_folder']
+    input_folder = step_folder_by_name['input_folder']
     for path, variable_definitions in variable_definitions_by_path.items():
         input_path = input_folder / path
         save_variable_data(input_path, data_by_id, variable_definitions)
     return batch_folder, batch_environment
 
 
-def _make_mode_folder_by_name(automation_folder, batch_folder):
-    mode_folder_by_name = _get_mode_folder_by_name(
+def _make_step_folder_by_name(automation_folder, batch_folder):
+    step_folder_by_name = _get_step_folder_by_name(
         automation_folder, batch_folder)
-    for folder in mode_folder_by_name.values():
+    for folder in step_folder_by_name.values():
         folder.mkdir(parents=True, exist_ok=True)
-    return mode_folder_by_name
+    return step_folder_by_name
 
 
-def _get_mode_folder_by_name(automation_folder, batch_folder):
+def _get_step_folder_by_name(automation_folder, batch_folder):
     return {
         _ + '_folder': automation_folder / batch_folder / _
-        for _ in MODE_NAMES}
+        for _ in STEP_NAMES}
 
 
 def _prepare_container_file_text(automation_definition):
@@ -591,10 +604,9 @@ def _prepare_container_file_text(automation_definition):
 
 
 def _prepare_script_environment(
-        mode_folder_by_name, custom_environment, with_path=False):
-    script_environment = {
-        'CROSSCOMPUTE_' + k.upper(): v for k, v in mode_folder_by_name.items()
-    } | custom_environment
+        step_folder_by_name, custom_environment, with_path=False):
+    script_environment = custom_environment | {
+        'CROSSCOMPUTE_' + k.upper(): v for k, v in step_folder_by_name.items()}
     if with_path:
         script_environment['PATH'] = getenv('PATH', '')
     return script_environment
@@ -620,23 +632,23 @@ def _prepare_batch_environment(
 
 
 def _process_batch(
-        automation_definition, batch_definition, mode_names,
-        extra_data_by_id_by_mode_name):
-    variable_data_by_id_by_mode_name = {}
+        automation_definition, batch_definition, step_names,
+        extra_data_by_id_by_step_name):
+    variable_data_by_id_by_step_name = {}
     automation_folder = automation_definition.folder
     batch_folder = batch_definition.folder
-    for mode_name in mode_names:
-        variable_data_by_id_by_mode_name[mode_name] = variable_data_by_id = {}
-        extra_data_by_id = extra_data_by_id_by_mode_name.get(mode_name, {})
-        mode_folder = automation_folder / batch_folder / mode_name
+    for step_name in step_names:
+        variable_data_by_id_by_step_name[step_name] = variable_data_by_id = {}
+        extra_data_by_id = extra_data_by_id_by_step_name.get(step_name, {})
+        step_folder = automation_folder / batch_folder / step_name
         variable_definitions = automation_definition.get_variable_definitions(
-            mode_name)
+            step_name)
         for variable_definition in variable_definitions:
             variable_id = variable_definition.id
             variable_path = variable_definition.path
             if variable_id in extra_data_by_id:
                 continue
-            path = mode_folder / variable_path
+            path = step_folder / variable_path
             try:
                 variable_data = process_variable_data(
                     path, variable_definition)
@@ -647,9 +659,9 @@ def _process_batch(
             variable_data_by_id[variable_id] = variable_data
         if extra_data_by_id:
             update_variable_data(
-                mode_folder / 'variables.dictionary', extra_data_by_id)
+                step_folder / 'variables.dictionary', extra_data_by_id)
             variable_data_by_id.update(extra_data_by_id)
-    return variable_data_by_id_by_mode_name
+    return variable_data_by_id_by_step_name
 
 
 def _get_image_name(automation_definition):
@@ -698,8 +710,8 @@ CONTAINER_PIPE_TEXT = (
     ' 2>>runs/next/debug/stderr.txt')
 CONTAINER_SCRIPT_NAME = '.run.sh'
 CONTAINER_ENV_NAME = '.run.env'
-CONTAINER_MODE_FOLDER_BY_NAME = {
-    _ + '_folder': 'runs/next/' + _ for _ in MODE_NAMES}
+CONTAINER_STEP_FOLDER_BY_NAME = {
+    _ + '_folder': 'runs/next/' + _ for _ in STEP_NAMES}
 
 
 L = getLogger(__name__)
