@@ -26,10 +26,13 @@ from ..constants import (
     AUTOMATION_ROUTE,
     AUTOMATION_VERSION,
     BATCH_ROUTE,
+    BUTTON_TEXT_BY_ID,
     DEBUG_VARIABLE_DICTIONARIES,
-    MODE_NAMES,
+    DESIGN_NAMES_BY_PAGE_ID,
+    INTERVAL_UNIT_NAMES,
+    PACKAGE_MANAGER_NAMES,
     PRINTER_BY_NAME,
-    RUN_ROUTE,
+    STEP_NAMES,
     STYLE_ROUTE,
     VARIABLE_ID_PATTERN,
     VARIABLE_ID_TEMPLATE_PATTERN,
@@ -94,35 +97,25 @@ class AutomationDefinition(Definition):
             validate_display_templates,
             validate_display_pages,
             validate_display_buttons,
-            validate_prints,
-        ]
+            validate_prints]
 
-    def get_variable_definitions(self, mode_name, with_all=False):
-        variable_definitions = self.variable_definitions_by_mode_name.get(
-            mode_name, [])
+    def get_variable_definitions(self, step_name, with_all=False):
+        variable_definitions = self.variable_definitions_by_step_name.get(
+            step_name, [])
         if with_all:
             variable_definitions = variable_definitions.copy()
-            for MODE_NAME in MODE_NAMES:
-                if mode_name == MODE_NAME:
+            for STEP_NAME in STEP_NAMES:
+                if step_name == STEP_NAME:
                     continue
                 variable_definitions.extend(self.get_variable_definitions(
-                    MODE_NAME))
+                    STEP_NAME))
         return variable_definitions
 
-    def get_template_path(self, template_id):
-        template_path_by_id = self.template_path_by_id
-        if template_id in template_path_by_id:
-            template_path = str(self.folder / template_path_by_id[template_id])
-        else:
-            template_path = f'crosscompute:templates/{template_id}.jinja2'
-        return template_path
-
-    def get_template_text(self, mode_name):
+    def get_template_text(self, step_name):
         automation_folder = self.folder
-        variable_definitions = self.get_variable_definitions(
-            mode_name)
-        template_definitions = self.template_definitions_by_mode_name[
-            mode_name]
+        variable_definitions = self.get_variable_definitions(step_name)
+        template_definitions = self.template_definitions_by_step_name[
+            step_name]
         return get_template_text(
             template_definitions, automation_folder, variable_definitions)
 
@@ -132,7 +125,7 @@ class AutomationDefinition(Definition):
             page_definition = self.page_definition_by_id[page_id]
             design_name = page_definition.configuration.get(
                 'design', design_name)
-        elif page_id in MODE_NAMES:
+        elif page_id in STEP_NAMES:
             variable_definitions = self.get_variable_definitions(page_id)
             if not variable_definitions:
                 design_name = 'none'
@@ -144,29 +137,26 @@ class AutomationDefinition(Definition):
         if button_id in button_definition_by_id:
             button_definition = button_definition_by_id[button_id]
             button_configuration = button_definition.configuration
-            button_text = button_configuration.get(
-                'button-text', button_text)
+            button_text = button_configuration.get('button-text', button_text)
         return button_text
 
 
 class VariableDefinition(Definition):
 
     def _initialize(self, kwargs):
-        self.mode_name = kwargs['mode_name']
+        self.step_name = kwargs['step_name']
         self._validation_functions = [
             validate_variable_identifiers,
-            validate_variable_configuration,
-        ]
+            validate_variable_configuration]
 
 
 class TemplateDefinition(Definition):
 
     def _initialize(self, kwargs):
         self.automation_folder = kwargs['automation_folder']
-        self.mode_name = kwargs.get('mode_name')
+        self.step_name = kwargs.get('step_name')
         self._validation_functions = [
-            validate_template_identifiers,
-        ]
+            validate_template_identifiers]
 
 
 class BatchDefinition(Definition):
@@ -176,8 +166,7 @@ class BatchDefinition(Definition):
         self.is_run = kwargs.get('is_run', False)
         self._validation_functions = [
             validate_batch_identifiers,
-            validate_batch_configuration,
-        ]
+            validate_batch_configuration]
 
     def get_status(self):
         status = Status.NEW
@@ -195,8 +184,7 @@ class DatasetDefinition(Definition):
         self.automation_folder = kwargs['automation_folder']
         self._validation_functions = [
             validate_dataset_identifiers,
-            validate_dataset_reference,
-        ]
+            validate_dataset_reference]
 
 
 class ScriptDefinition(Definition):
@@ -204,8 +192,7 @@ class ScriptDefinition(Definition):
     def _initialize(self, kwargs):
         self.automation_folder = kwargs['automation_folder']
         self._validation_functions = [
-            validate_script_identifiers,
-        ]
+            validate_script_identifiers]
 
     def get_command_string(self):
         command_string = self.command
@@ -246,24 +233,21 @@ class PackageDefinition(Definition):
 
     def _initialize(self, kwargs):
         self._validation_functions = [
-            validate_package_identifiers,
-        ]
+            validate_package_identifiers]
 
 
 class PortDefinition(Definition):
 
     def _initialize(self, kwargs):
         self._validation_functions = [
-            validate_port_identifiers,
-        ]
+            validate_port_identifiers]
 
 
 class StyleDefinition(Definition):
 
     def _initialize(self, kwargs):
         self._validation_functions = [
-            validate_style_identifiers,
-        ]
+            validate_style_identifiers]
 
 
 class PageDefinition(Definition):
@@ -271,8 +255,7 @@ class PageDefinition(Definition):
     def _initialize(self, kwargs):
         self._validation_functions = [
             validate_page_identifiers,
-            validate_page_configuration,
-        ]
+            validate_page_configuration]
 
 
 class ButtonDefinition(Definition):
@@ -280,8 +263,7 @@ class ButtonDefinition(Definition):
     def _initialize(self, kwargs):
         self._validation_functions = [
             validate_button_identifiers,
-            validate_button_configuration,
-        ]
+            validate_button_configuration]
 
 
 class TokenDefinition(Definition):
@@ -289,24 +271,21 @@ class TokenDefinition(Definition):
     def _initialize(self, kwargs):
         self.automation_folder = kwargs['automation_folder']
         self._validation_functions = [
-            validate_token_identifiers,
-        ]
+            validate_token_identifiers]
 
 
 class GroupDefinition(Definition):
 
     def _initialize(self, kwargs):
         self._validation_functions = [
-            validate_group_identifiers,
-        ]
+            validate_group_identifiers]
 
 
 class PermissionDefinition(Definition):
 
     def _initialize(self, kwargs):
         self._validation_functions = [
-            validate_permission_identifiers,
-        ]
+            validate_permission_identifiers]
 
 
 class PrintDefinition(Definition):
@@ -316,8 +295,7 @@ class PrintDefinition(Definition):
             validate_print_identifiers,
             validate_print_configuration,
             validate_header_footer_options,
-            validate_page_number_options,
-        ]
+            validate_page_number_options]
 
 
 def save_raw_configuration(configuration_path, configuration):
@@ -421,8 +399,7 @@ def validate_automation_identifiers(configuration):
         'description': description,
         'version': version,
         'slug': slug,
-        'uri': uri,
-    }
+        'uri': uri}
 
 
 def validate_imports(configuration):
@@ -454,32 +431,29 @@ def validate_imports(configuration):
         _.slug for _ in automation_definitions], 'duplicate slug {x}')
     return {
         'import_configurations': import_configurations,
-        'automation_definitions': automation_definitions,
-    }
+        'automation_definitions': automation_definitions}
 
 
 def validate_variables(configuration):
-    variable_definitions_by_mode_name = {}
+    variable_definitions_by_step_name = {}
     view_names = set()
-    for mode_name in MODE_NAMES:
-        mode_configuration = get_dictionary(configuration, mode_name)
+    for step_name in STEP_NAMES:
+        step_configuration = get_dictionary(configuration, step_name)
         variable_dictionaries = get_dictionaries(
-            mode_configuration, 'variables')
-        if mode_name == 'debug':
+            step_configuration, 'variables')
+        if step_name == 'debug':
             variable_dictionaries[:0] = DEBUG_VARIABLE_DICTIONARIES
         variable_definitions = [VariableDefinition(
-            _, mode_name=mode_name,
-        ) for _ in variable_dictionaries]
+            _, step_name=step_name) for _ in variable_dictionaries]
         assert_unique_values([
             _.id for _ in variable_definitions
-        ], f'duplicate variable id {{x}} in {mode_name}')
-        variable_definitions_by_mode_name[mode_name] = variable_definitions
+        ], f'duplicate variable id {{x}} in {step_name}')
+        variable_definitions_by_step_name[step_name] = variable_definitions
         view_names.update(_.view_name for _ in variable_definitions)
     L.debug('view_names = %s', list(view_names))
     return {
-        'variable_definitions_by_mode_name': variable_definitions_by_mode_name,
-        '___view_names': view_names,
-    }
+        'variable_definitions_by_step_name': variable_definitions_by_step_name,
+        '___view_names': view_names}
 
 
 def validate_variable_views(configuration):
@@ -499,20 +473,19 @@ def validate_variable_views(configuration):
 
 
 def validate_templates(configuration):
-    template_definitions_by_mode_name = {}
+    template_definitions_by_step_name = {}
     automation_folder = configuration.folder
-    for mode_name in MODE_NAMES:
-        mode_configuration = get_dictionary(configuration, mode_name)
+    for step_name in STEP_NAMES:
+        step_configuration = get_dictionary(configuration, step_name)
         template_definitions = [TemplateDefinition(
-            _, automation_folder=automation_folder, mode_name=mode_name,
-        ) for _ in get_dictionaries(mode_configuration, 'templates')]
+            _, automation_folder=automation_folder, step_name=step_name,
+        ) for _ in get_dictionaries(step_configuration, 'templates')]
         assert_unique_values([
             _.id for _ in template_definitions],
-            f'duplicate template id {{x}} in {mode_name}')
-        template_definitions_by_mode_name[mode_name] = template_definitions
+            f'duplicate template id {{x}} in {step_name}')
+        template_definitions_by_step_name[step_name] = template_definitions
     return {
-        'template_definitions_by_mode_name': template_definitions_by_mode_name,
-    }
+        'template_definitions_by_step_name': template_definitions_by_step_name}
 
 
 def validate_batches(configuration):
@@ -532,20 +505,14 @@ def validate_batches(configuration):
         _.name for _ in batch_definitions], 'duplicate batch name {x}')
     assert_unique_values([
         _.uri for _ in batch_definitions], 'duplicate batch uri {x}')
-    return {
-        'batch_definitions': batch_definitions,
-        'run_definitions': [],
-    }
+    return {'batch_definitions': batch_definitions}
 
 
 def validate_environment(configuration):
     d = get_dictionary(configuration, 'environment')
-    engine_name = get_engine_name(d)
-    parent_image_name = d.get('image', 'python').strip()
-    package_definitions = [PackageDefinition(_) for _ in get_dictionaries(
-        d, 'packages')]
     port_definitions = get_port_definitions(
-        d, configuration.get_variable_definitions('log', with_all=True))
+        d, configuration.get_variable_definitions('log')
+        + configuration.get_variable_definitions('debug'))
     environment_variable_ids = get_environment_variable_ids(get_dictionaries(
         d, 'variables'))
     batch_concurrency_name = d.get('batch', 'process').lower()
@@ -555,14 +522,14 @@ def validate_environment(configuration):
     interval_text = d.get('interval', '').strip()
     interval_timedelta = get_interval_timedelta(interval_text)
     return {
-        'engine_name': engine_name,
-        'parent_image_name': parent_image_name,
-        'package_definitions': package_definitions,
+        'engine_name': get_engine_name(d),
+        'parent_image_name': d.get('image', 'python').strip(),
+        'package_definitions': [PackageDefinition(_) for _ in get_dictionaries(
+            d, 'packages')],
         'port_definitions': port_definitions,
         'environment_variable_ids': environment_variable_ids,
         'batch_concurrency_name': batch_concurrency_name,
-        'interval_timedelta': interval_timedelta,
-    }
+        'interval_timedelta': interval_timedelta}
 
 
 def validate_datasets(configuration):
@@ -570,9 +537,7 @@ def validate_datasets(configuration):
     dataset_definitions = [DatasetDefinition(
         _, automation_folder=automation_folder,
     ) for _ in get_dictionaries(configuration, 'datasets')]
-    return {
-        'dataset_definitions': dataset_definitions,
-    }
+    return {'dataset_definitions': dataset_definitions}
 
 
 def validate_scripts(configuration):
@@ -580,9 +545,7 @@ def validate_scripts(configuration):
     script_definitions = [ScriptDefinition(
         _, automation_folder=automation_folder,
     ) for _ in get_dictionaries(configuration, 'scripts')]
-    return {
-        'script_definitions': script_definitions,
-    }
+    return {'script_definitions': script_definitions}
 
 
 def validate_display_styles(configuration):
@@ -592,8 +555,7 @@ def validate_display_styles(configuration):
     automation_uri = configuration.uri
     reference_time = time()
     style_definitions = []
-    for raw_style_definition in get_dictionaries(
-            display_dictionary, 'styles'):
+    for raw_style_definition in get_dictionaries(display_dictionary, 'styles'):
         style_definition = StyleDefinition(raw_style_definition)
         style_uri = style_definition.uri
         if '//' not in style_uri:
@@ -612,8 +574,7 @@ def validate_display_styles(configuration):
         style_definitions.append(style_definition)
     return {
         'style_definitions': style_definitions,
-        'css_uris': [_.uri for _ in style_definitions],
-    }
+        'css_uris': [_.uri for _ in style_definitions]}
 
 
 def validate_display_templates(configuration):
@@ -626,9 +587,7 @@ def validate_display_templates(configuration):
             raw_template_definition, automation_folder=automation_folder)
         template_id = template_definition.id
         template_path_by_id[template_id] = template_definition.path
-    return {
-        'template_path_by_id': template_path_by_id,
-    }
+    return {'template_path_by_id': template_path_by_id}
 
 
 def validate_display_pages(configuration):
@@ -636,9 +595,7 @@ def validate_display_pages(configuration):
     page_definitions = [PageDefinition(_) for _ in get_dictionaries(
         display_dictionary, 'pages')]
     page_definition_by_id = {_.id: _ for _ in page_definitions}
-    return {
-        'page_definition_by_id': page_definition_by_id,
-    }
+    return {'page_definition_by_id': page_definition_by_id}
 
 
 def validate_display_buttons(configuration):
@@ -646,9 +603,7 @@ def validate_display_buttons(configuration):
     button_definitions = [ButtonDefinition(_) for _ in get_dictionaries(
         display_dictionary, 'buttons')]
     button_definition_by_id = {_.id: _ for _ in button_definitions}
-    return {
-        'button_definition_by_id': button_definition_by_id,
-    }
+    return {'button_definition_by_id': button_definition_by_id}
 
 
 def validate_authorization(configuration):
@@ -665,16 +620,13 @@ def validate_authorization(configuration):
     group_definitions = child_group_definitions + parent_group_definitions
     return {
         'identities_by_token': identities_by_token,
-        'group_definitions': group_definitions,
-    }
+        'group_definitions': group_definitions}
 
 
 def validate_prints(configuration):
     print_definitions = [PrintDefinition(_) for _ in get_dictionaries(
         configuration, 'prints')]
-    return {
-        'print_definitions': print_definitions,
-    }
+    return {'print_definitions': print_definitions}
 
 
 def validate_template_identifiers(template_dictionary):
@@ -689,8 +641,7 @@ def validate_template_identifiers(template_dictionary):
             f'could not find template {template_path}')
     return {
         'id': template_dictionary.get('id', template_path.stem),
-        'path': template_path,
-    }
+        'path': template_path}
 
 
 def validate_variable_identifiers(variable_dictionary):
@@ -711,16 +662,13 @@ def validate_variable_identifiers(variable_dictionary):
     return {
         'id': variable_id,
         'view_name': view_name,
-        'path': Path(variable_path),
-    }
+        'path': Path(variable_path)}
 
 
 def validate_variable_configuration(variable_dictionary):
     variable_configuration = get_dictionary(
         variable_dictionary, 'configuration')
-    return {
-        'configuration': variable_configuration,
-    }
+    return {'configuration': variable_configuration}
 
 
 def validate_batch_identifiers(batch_dictionary):
@@ -751,11 +699,9 @@ def validate_batch_identifiers(batch_dictionary):
             if k in batch_dictionary:
                 batch_dictionary[k] = v
     if data_by_id is not None:
-        if is_run:
-            uri = RUN_ROUTE.format(run_slug=slug)
-        else:
+        if not is_run:
             slug = format_slug(slug)
-            uri = BATCH_ROUTE.format(batch_slug=slug)
+        uri = BATCH_ROUTE.format(batch_slug=slug)
         d.update({'slug': slug, 'uri': uri})
     return d
 
@@ -765,14 +711,11 @@ def validate_batch_configuration(batch_dictionary):
     batch_configuration = get_dictionary(batch_dictionary, 'configuration')
     return {
         'reference': batch_reference,
-        'configuration': batch_configuration,
-    }
+        'configuration': batch_configuration}
 
 
 def validate_dataset_identifiers(dataset_dictionary):
-    return {
-        'path': get_folder_plus_path(dataset_dictionary),
-    }
+    return {'path': get_folder_plus_path(dataset_dictionary)}
 
 
 def validate_dataset_reference(dataset_dictionary):
@@ -780,17 +723,22 @@ def validate_dataset_reference(dataset_dictionary):
     dataset_reference = get_dictionary(dataset_dictionary, 'reference')
     reference_path = get_folder_plus_path(dataset_reference)
     if reference_path:
-        if not (automation_folder / reference_path).exists():
-            raise CrossComputeConfigurationError(
-                f'could not find dataset reference {reference_path}')
+        source_path = automation_folder / reference_path
+        if not source_path.exists():
+            if source_path.is_symlink():
+                raise CrossComputeConfigurationError(
+                    f'invalid symlink for dataset reference {reference_path}')
+            elif source_path.name == 'runs':
+                source_path.mkdir(parents=True)
+            else:
+                raise CrossComputeConfigurationError(
+                    f'could not find dataset reference {reference_path}')
         target_path = dataset_dictionary.path
         if target_path.exists() and not target_path.is_symlink():
             raise CrossComputeConfigurationError(
                 'refusing to overwrite existing dataset; please delete '
                 f'{target_path} from the disk as defined')
-    return {
-        'reference': dataset_reference,
-    }
+    return {'reference': dataset_reference}
 
 
 def validate_script_identifiers(script_dictionary):
@@ -813,8 +761,7 @@ def validate_script_identifiers(script_dictionary):
     return {
         'folder': Path(folder),
         'path': path,
-        'command': command,
-    }
+        'command': command}
 
 
 def validate_package_identifiers(package_dictionary):
@@ -828,8 +775,7 @@ def validate_package_identifiers(package_dictionary):
             f'"{manager_name}" manager is not supported')
     return {
         'id': package_id,
-        'manager_name': manager_name,
-    }
+        'manager_name': manager_name}
 
 
 def validate_port_identifiers(port_dictionary):
@@ -845,8 +791,7 @@ def validate_port_identifiers(port_dictionary):
             f'{port_number} must be an integer')
     return {
         'id': port_id,
-        'number': port_number,
-    }
+        'number': port_number}
 
 
 def validate_style_identifiers(style_dictionary):
@@ -855,10 +800,7 @@ def validate_style_identifiers(style_dictionary):
     if not uri and not path:
         raise CrossComputeConfigurationError(
             'uri or path required for each style')
-    return {
-        'uri': uri,
-        'path': Path(path),
-    }
+    return {'uri': uri, 'path': Path(path)}
 
 
 def validate_page_identifiers(page_dictionary):
@@ -869,9 +811,7 @@ def validate_page_identifiers(page_dictionary):
     if page_id not in DESIGN_NAMES_BY_PAGE_ID:
         raise CrossComputeConfigurationError(
             f'{page_id} page not supported for page configuration')
-    return {
-        'id': page_id,
-    }
+    return {'id': page_id}
 
 
 def validate_page_configuration(page_dictionary):
@@ -882,9 +822,7 @@ def validate_page_configuration(page_dictionary):
     if design_name not in design_names:
         raise CrossComputeConfigurationError(
             f'"{design_name}" design not supported for {page_id} page')
-    return {
-        'configuration': page_configuration,
-    }
+    return {'configuration': page_configuration}
 
 
 def validate_button_identifiers(button_dictionary):
@@ -895,16 +833,12 @@ def validate_button_identifiers(button_dictionary):
     if button_id not in BUTTON_TEXT_BY_ID:
         raise CrossComputeConfigurationError(
             f'{button_id} button not supported for button configuration')
-    return {
-        'id': button_id,
-    }
+    return {'id': button_id}
 
 
 def validate_button_configuration(button_dictionary):
     button_configuration = get_dictionary(button_dictionary, 'configuration')
-    return {
-        'configuration': button_configuration,
-    }
+    return {'configuration': button_configuration}
 
 
 def validate_token_identifiers(token_dictionary):
@@ -921,6 +855,7 @@ def validate_token_identifiers(token_dictionary):
             f'{suffix} not supported for token paths')
     identities_by_token = {}
     for token, identities in d.items():
+        token = str(token)
         variable_match = VARIABLE_ID_TEMPLATE_PATTERN.match(token)
         if variable_match:
             variable_id = variable_match.group(1)
@@ -943,9 +878,8 @@ def validate_group_identifiers(group_dictionary):
     group_permissions = [PermissionDefinition(_) for _ in get_dictionaries(
         group_dictionary, 'permissions')]
     return {
-        'configuration': group_configuration,
-        'permissions': group_permissions,
-    }
+        'configuration': group_configuration or {},
+        'permissions': group_permissions}
 
 
 def validate_permission_identifiers(permission_dictionary):
@@ -961,10 +895,7 @@ def validate_permission_identifiers(permission_dictionary):
     if permission_action not in PERMISSION_ACTIONS:
         raise CrossComputeConfigurationError(
             f'"{permission_action}" action not supported')
-    return {
-        'id': permission_id,
-        'action': permission_action,
-    }
+    return {'id': permission_id, 'action': permission_action}
 
 
 def validate_print_identifiers(print_dictionary):
@@ -986,16 +917,13 @@ def validate_print_identifiers(print_dictionary):
     return {
         'format': print_format,
         'folder': print_folder,
-        'name': print_name,
-    }
+        'name': print_name}
 
 
 def validate_print_configuration(print_dictionary):
     print_configuration = get_dictionary(
         print_dictionary, 'configuration')
-    return {
-        'configuration': print_configuration,
-    }
+    return {'configuration': print_configuration}
 
 
 def validate_header_footer_options(print_dictionary):
@@ -1079,11 +1007,11 @@ def get_port_definitions(environment_dictionary, variable_definitions):
         except StopIteration:
             raise CrossComputeConfigurationError(
                 f'{port_id} port must have a matching variable definition')
-        mode_name = variable_definition.mode_name
-        if mode_name not in ['log', 'debug']:
+        step_name = variable_definition.step_name
+        if step_name not in ['log', 'debug']:
             raise CrossComputeConfigurationError(
                 f'{port_id} port must correspond to a log or debug variable')
-        port_definition.mode_name = mode_name
+        port_definition.step_name = step_name
     return port_definitions
 
 
@@ -1212,31 +1140,6 @@ def assert_unique_values(xs, message):
             raise CrossComputeConfigurationError(message.format(x=x))
 
 
-PACKAGE_MANAGER_NAMES = [
-    'apt',
-    'dnf',
-    'npm',
-    'pip',
-]
-
-
-DESIGN_NAMES_BY_PAGE_ID = {
-    'automation': ['input', 'output', 'none'],
-    'input': ['flex-vertical', 'none'],
-    'output': ['flex-vertical', 'none'],
-    'log': ['flex-vertical', 'none'],
-    'debug': ['flex-vertical', 'none'],
-}
-
-
-BUTTON_TEXT_BY_ID = {
-    'run': 'Run',
-}
-
-
-INTERVAL_UNIT_NAMES = 'seconds', 'minutes', 'hours', 'days', 'weeks'
-
-
 RUN_PY = Template('''\
 import inspect
 from os import getenv
@@ -1249,7 +1152,7 @@ folder_by_name = {}
 for x in 'input_folder', 'output_folder', 'log_folder', 'debug_folder':
     folder_by_name[x] = Path(getenv('CROSSCOMPUTE_' + x.upper()))
 d = {}
-for x in inspect.getargspec(run.plot).args:
+for x in inspect.getargspec($function_string).args:
     d[x] = folder_by_name[x]
 $function_string(**d)''')
 
