@@ -1,46 +1,56 @@
+from logging import getLogger
 from pathlib import Path
 from time import time
 
 from invisibleroads_macros_disk import is_path_in_folder
 
 from ..constants import (
-    MODE_CODE_BY_NAME,
-    MODE_ROUTE,
-    RUN_ROUTE)
+    BATCH_ROUTE,
+    STEP_CODE_BY_NAME,
+    STEP_ROUTE)
 
 
 class DiskDatabase():
 
-    def __init__(self, configuration, infos_by_timestamp):
+    def __init__(self, configuration, changes):
         self._configuration = configuration
-        self._infos_by_timestamp = infos_by_timestamp
+        self._changes = changes
         self._memory = learn(configuration)
 
     def grok(self, paths):
         changed_infos = []
+        variable_ids = []
         for path in paths:
+            path = Path(path).resolve()
+            if path.is_dir():
+                continue
             try:
                 infos = self._get(path)
             except KeyError:
                 continue
-            changed_infos.extend(infos)
+            for info in infos:
+                if info['code'] == 'v':
+                    variable_id = info['id']
+                    if info['id'] in variable_ids:
+                        continue
+                    variable_ids.append(variable_id)
+                changed_infos.append(info)
         if changed_infos:
-            self._infos_by_timestamp[time()] = changed_infos
+            self._changes[time()] = changed_infos
         return changed_infos
 
     def _get(self, path):
         configuration = self._configuration
-        path = Path(path).resolve()
         for automation_definition in configuration.automation_definitions:
             automation_folder = automation_definition.folder
             runs_folder = automation_folder / 'runs'
             if not is_path_in_folder(path, runs_folder):
                 continue
             run_id = path.relative_to(runs_folder).parts[0]
-            run_uri = RUN_ROUTE.format(run_slug=run_id)
+            batch_uri = BATCH_ROUTE.format(batch_slug=run_id)
             memory = DiskMemory()
             add_variable_infos_from_folder(
-                memory, automation_definition, runs_folder / run_id, run_uri)
+                memory, automation_definition, runs_folder / run_id, batch_uri)
             infos = memory.get(path)
             break
         else:
@@ -109,15 +119,15 @@ def add_variable_infos_from_folder(
     info = {'code': 'v'}
     automation_uri = automation_definition.uri
     uri = automation_uri + batch_uri
-    d = automation_definition.variable_definitions_by_mode_name
-    for mode_name, variable_definitions in d.items():
-        mode_code = MODE_CODE_BY_NAME[mode_name]
-        mode_uri = MODE_ROUTE.format(mode_code=mode_code)
-        folder = absolute_batch_folder / mode_name
+    d = automation_definition.variable_definitions_by_step_name
+    for step_name, variable_definitions in d.items():
+        step_code = STEP_CODE_BY_NAME[step_name]
+        step_uri = STEP_ROUTE.format(step_code=step_code)
+        folder = absolute_batch_folder / step_name
         for variable_definition in variable_definitions:
             variable_id = variable_definition.id
             if variable_id != 'return_code':
-                info_uri = uri + mode_uri
+                info_uri = uri + step_uri
             else:
                 info_uri = uri
             variable_info = info | {'id': variable_id, 'uri': info_uri}
@@ -135,8 +145,8 @@ def add_template_infos(memory, configuration):
         automation_folder = automation_definition.folder
         automation_uri = automation_definition.uri
         template_info = info | {'uri': automation_uri}
-        d = automation_definition.template_definitions_by_mode_name
-        for mode_name, template_definitions in d.items():
+        d = automation_definition.template_definitions_by_step_name
+        for step_name, template_definitions in d.items():
             for template_definition in template_definitions:
                 if 'path' not in template_definition:
                     continue
@@ -158,3 +168,6 @@ def add_style_infos(memory, configuration):
                 continue
             path = automation_folder / style_definition['path']
             memory.add(path, info)
+
+
+L = getLogger(__name__)
