@@ -3,7 +3,6 @@ from logging import getLogger
 
 from invisibleroads_macros_disk import make_folder, make_random_folder
 from invisibleroads_macros_log import get_timestamp, LONGSTAMP_TEMPLATE
-from invisibleroads_macros_process import StoppableProcess
 from invisibleroads_macros_web.port import find_open_port
 
 from crosscompute.constants import (
@@ -30,6 +29,8 @@ from crosscompute.scripts.serve import (
     configure_argument_parser_for_serving,
     configure_serving_from,
     serve_with)
+from crosscompute.settings import (
+    StoppableProcess)
 
 
 def do(arguments=None):
@@ -69,14 +70,14 @@ def print_with(automation, args):
     for automation_definition in get_selected_automation_definitions(
             automation.definitions):
         # TODO: Consider using ds returned from run_automation
-        run_automation(automation_definition, with_rebuild=True)
+        run_automation(
+            automation_definition, args.environment, with_rebuild=True)
         for print_definition in automation_definition.print_definitions:
             batch_dictionaries = get_batch_dictionaries(
                 automation_definition, print_definition, timestamp)
             print_packs.append((print_definition, batch_dictionaries))
     args.port = port
     args.with_browser = False
-    args.with_refresh = False
     args.with_restart = False
     server_process = StoppableProcess(
         name='serve', target=serve_with, args=(automation, args))
@@ -86,6 +87,10 @@ def print_with(automation, args):
             Printer = PRINTER_BY_NAME[print_definition.format]
             printer = Printer(f'http://127.0.0.1:{port}{args.root_uri}')
             printer.render(batch_dictionaries, print_definition)
+    except KeyboardInterrupt:
+        pass
+    except Exception as e:
+        L.exception(e)
     finally:
         server_process.stop()
 
